#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" This package implement a Web Site Cloner 
and his HTTP server to launch it. """

###################
#    This package implement a Web Site Cloner and his HTTP server to launch it.
#    Copyright (C) 2021  Maurice Lambert

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
###################

from sys import argv
from os import path

try:
    from .WebSiteCloner import main as cloner
    from .HTTPLogger import main as server
except ImportError:
    from WebSiteCloner import main as cloner
    from HTTPLogger import main as server

if len(argv) > 1:
    function = argv.pop(1).lower()
    if function == "cloner":
        cloner()
        exit(0)
    elif function == "server":
        server()
        exit(0)

print(
    """
WebSiteCloner  Copyright (C) 2021  Maurice Lambert
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.
"""
)

print(f"""USAGES: 
    {path.basename(argv[0])} server
    {path.basename(argv[0])} cloner
HELP:
    {path.basename(argv[0])} cloner --help
    {path.basename(argv[0])} server -h
""")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" This package implement a Web Site Cloner 
and his HTTP server to launch it. """

###################
#    This package implement a Web Site Cloner and his HTTP server to launch it.
#    Copyright (C) 2021  Maurice Lambert

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
###################

__version__ = "0.0.1"
__all__ = ["WebSiteCloner", "cloner", "server", "HTTPLogger"]

from .WebSiteCloner import WebSiteCloner, main as cloner
from .HTTPLogger import HTTPLogger, main as server

print(
    """
WebSiteCloner  Copyright (C) 2021  Maurice Lambert
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.
"""
)



from setuptools import setup, find_packages
import WebSiteCloner

setup(
    name = 'WebSiteCloner',
 
    version = WebSiteCloner.__version__,
    packages = find_packages(),
    install_requires = [],

    author = "Maurice Lambert", 
    author_email = "mauricelambert434@gmail.com",
 
    description = "This package implement a Web Site Cloner and his HTTP server to launch it.",
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
 
    include_package_data = True,

    url = 'https://github.com/mauricelambert/WebSiteCloner',
 
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
    ],
 
    entry_points = {
        'console_scripts': [
            'WebSiteCloner = WebSiteCloner:cloner',
            'WebClonerServer = WebSiteCloner:server'
        ],
    },
    python_requires='>=3.6',
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" This script implement a HTTP Logger class to launch
cloned WebSite and save the requests. """

###################
#    This script implement a HTTP Logger class to launch cloned WebSite and save the requests.
#    Copyright (C) 2021  Maurice Lambert

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
###################

from http.server import HTTPServer, SimpleHTTPRequestHandler
from argparse import ArgumentParser
from os import path, chdir
import logging


class HTTPLogger(SimpleHTTPRequestHandler):

    """This class is basic HTTP Server to launch a cloned WebSite.
    This class log the request and data sending with it (credentials, secret files...)."""

    def do_GET(self, data=None) -> None:

        """ This function log HTTP GET request and return the file. """

        self.sys_version = ""
        self.server_version = CONSTANTES.HEADER_SERVER

        logging.info(
            f'{self.client_address} asking "{self.command} {self.path}" on {self.headers.get("Host")}'
        )

        headers_string = ""
        for header, value in self.headers.items():
            headers_string += f"{header}:{value}; "

        logging.info(f"Headers sending by {self.client_address}: {headers_string}")

        if data:
            logging.info(f"Data sending by {self.client_address}: {data}")

        SimpleHTTPRequestHandler.do_GET(self)

    def do_POST(self) -> None:

        """ This function get data from HTTP POST request. """

        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length)

        self.do_GET(data=post_data)

    def end_headers(self) -> None:

    	""" This function send custom headers and call the real end_headers function. """

    	for header, value in CONSTANTES.HEADERS:
    		self.send_header(header, value)

    	SimpleHTTPRequestHandler.end_headers(self)


class CONSTANTES:

    """ Constantes class for permanent HTTP server values. """

    HEADER_SERVER = "WebSiteCloner"
    LOGFILE = None
    LOGLEVEL = logging.DEBUG
    DIRECTORY = "CloneWebSite"
    HEADERS = []
    PORT = 80
    INTERFACE = "0.0.0.0"


def parse():
    parser = ArgumentParser()
    parser.add_argument(
        "domain",
        help="Domain cloned or directory.",
    )
    parser.add_argument("--logfile", "-f", help="HTTP server logs file.", default=None)
    parser.add_argument(
        "--headers",
        "-H",
        help="Additionnal headers for response with this format: name:value,name:value... (Server:Apache).",
        default=None,
    )
    parser.add_argument(
        "--server", "-S", help="Server header.", default="WebSiteCloner"
    )
    parser.add_argument(
        "--port", "-P", help="Port to launch server.", default=80, type=int
    )
    parser.add_argument(
        "--interface", "-I", help="Interface to launch server.", default="0.0.0.0"
    )
    return parser.parse_args()


def main():
    args = parse()

    if args.headers:
        for header in args.headers.split(","):
            CONSTANTES.HEADERS.append(tuple(header.split(":", 1)))

    CONSTANTES.HEADER_SERVER = args.server
    CONSTANTES.PORT = args.port
    CONSTANTES.INTERFACE = args.interface
    CONSTANTES.LOGFILE = args.logfile

    build_path = path.join(CONSTANTES.DIRECTORY, args.domain)

    if path.isdir(args.domain):
        chdir(args.domain)
        CONSTANTES.DIRECTORY = args.domain
    elif path.isdir(build_path):
        chdir(build_path)
        CONSTANTES.DIRECTORY = build_path
    else:
        logging.critical(f"{args.domain} and {build_path} are not a directory.")
        exit(2)

    logging.basicConfig(
        format='[%(asctime)s] "%(message)s"',
        datefmt="%m/%d/%Y %H:%M:%S",
        level=CONSTANTES.LOGLEVEL,
        filename=CONSTANTES.LOGFILE,
    )

    with HTTPServer((CONSTANTES.INTERFACE, CONSTANTES.PORT), HTTPLogger) as httpd:
        logging.info(f"Server is running in : {CONSTANTES.DIRECTORY}")
        logging.info(f"Server is running on : {CONSTANTES.INTERFACE}:{CONSTANTES.PORT}")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            exit(0)


if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" This script implement a the WebSite Cloner with
GetURLsFromHTML class and WebSiteCloner class. """

###################
#    This script implement a the WebSite Cloner with GetURLsFromHTML class and WebSiteCloner class.
#    Copyright (C) 2021  Maurice Lambert

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
###################

from urllib.parse import urlparse, ParseResult
from urllib.error import URLError, HTTPError
from http.client import HTTPResponse
from argparse import ArgumentParser
from html.parser import HTMLParser
from urllib.request import urlopen
from typing import List, Tuple
from os import path, makedirs
from re import match
import logging


class UrlError(ValueError):
    pass


class GetURLsFromHTML(HTMLParser):

    """ This class get URLs from HTML. """

    def __init__(self, master):
        super().__init__()
        self.master = master

    def handle_starttag(self, tag: str, attributes: List[Tuple[str, str]]) -> None:

        """ This function get URLs attributes and values. """

        for attribute in attributes:
            if attribute[0] in [
                "src",
                "href",
                "cite",
                "code",
                "codebase",
                "data",
                "poster",
                "srcset",
                "srcdoc",
            ]:
                if not match(
                    "^https?://(w{0,3})[.]?(([a-zA-Z-]+)[.])+[a-zA-Z]{0,5}[!\x23-\x3B=\x3F-\x5B\]_\x61-\x7A~]*$",
                    attribute[1],
                ) or match(
                    "^https?://(w{1,3})?[.]?" + self.master.urlparse.netloc + ".*$",
                    attribute[1],
                ):
                    self.master.add_new_url(attribute[1])
                else:
                    return


class WebSiteCloner:

    """ This class Clone a Page and her ressources. """

    def __init__(
        self,
        url: str,
        directory: str = "CloneWebSite",
        recursive: bool = False,
        replace_domain: str = None,
        loglevel: int = logging.INFO,
        logfile: str = None,
        replace_scheme: str = None,
    ):
        self.urlparse: ParseResult = urlparse(url)
        self.html: bytes = None
        self.html_parser: GetURLsFromHTML = GetURLsFromHTML(self)
        self.directory = directory
        self.recursive = recursive
        self.replace_domain = replace_domain.encode() if replace_domain else None
        self.urls_to_parse: List[ParseResult] = [self.urlparse.path]
        self.urls_parsed: List[ParseResult] = []
        self.loglevel = loglevel
        self.logfile = logfile
        self.replace_scheme = replace_scheme.encode() if replace_scheme else None

    def launch(self) -> None:

        """ Launcher to copy a website. """

        logging.basicConfig(
            format="%(asctime)s %(levelname)s : %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=self.loglevel,
            filename=self.logfile,
        )

        first = True

        while self.urls_to_parse:
            url = self.urls_to_parse.pop()
            self.urls_parsed.append(url)

            response = self.get_data(url)

            if not response:
                continue

            is_html = "text/html" in response.getheader("content-type")
            data = response.read()

            self.write_file(
                url,
                data,
            )

            if is_html and (self.recursive or first):
                self.html_parser.feed(data.decode())
                if first:
                    first = False
            elif first and not is_html:
                logging.critical(f"URL {self.url} is not valid in this context.")
                raise UrlError(f"URL {self.url} is not valid in this context.")

    def add_new_url(self, url: str) -> None:

        """ This function add new url to parse. """

        url_parsed = urlparse(url)

        if (
            (url_parsed.netloc == self.urlparse.netloc or url_parsed.netloc == "")
            and url_parsed.path not in self.urls_to_parse
            and url_parsed.path not in self.urls_parsed
        ):
            self.urls_to_parse.append(url_parsed.path)

    def get_complete_url(self, url: str) -> str:

        """ This function build a complete url. """

        if url and url[0] == "/":
            complete_url = "https://" + self.urlparse.netloc + url
        elif self.urlparse.path != "/":
            complete_url = (
                "https://" + self.urlparse.netloc + self.urlparse.path + "/" + url
            )
        else:
            complete_url = "https://" + self.urlparse.netloc + self.urlparse.path + url

        return complete_url

    def get_data(self, url: str) -> HTTPResponse:

        """ This function return HTTP response. """

        url = self.get_complete_url(url)

        try:
            response = urlopen(url)
        except URLError:
            logging.error(
                f"This URL isn't valid. Failed to get ressources from this url: {url}"
            )
            return None
        except HTTPError as error:
            logging.error(
                f"HTTP {error.code} error. Failed to get ressources from this url : {url}"
            )
            return None

        logging.info(f"Get ressources from this URL : {url}")
        return response

    def write_file(self, url: str, data: bytes) -> bytes:

        """ This function get URL path and write the file in the good location. """

        url_parsed = urlparse(url)
        directory = self.get_directory_from_url(url, url_parsed)
        full_path = self.get_full_path(url, directory, url_parsed)

        if self.replace_domain:
            data = data.replace(self.urlparse.netloc.encode(), self.replace_domain)
        if self.replace_scheme:
            data = data.replace(self.urlparse.scheme.encode(), self.replace_scheme)

        file = open(full_path, "wb")
        file.write(data)
        file.close()

    def get_full_path(self, url: str, directory: str, url_parsed: ParseResult) -> str:

        """ This function return full path (directory + filename) to write the file. """

        filename = path.basename(url_parsed.path)
        if "." not in filename:
            directory = path.join(directory, filename)
            filename = "index.html"

        logging.debug(f"Path for URL: {url} (path: {url_parsed.path}) is {directory}")

        full_path = path.join(self.urlparse.netloc, directory)
        if path.isdir(self.directory):
            full_path = path.join(self.directory, full_path)
        if not path.exists(full_path):
            makedirs(full_path)

        logging.debug(
            f"Finally path for URL: {url} (path: {url_parsed.path}) is {full_path}"
        )

        full_path = path.join(full_path, filename)
        if path.exists(full_path):
            if path.isdir(full_path):
                full_path = path.join(full_path, "index.html")

        logging.debug(
            f"The filename for URL: {url} (path: {url_parsed.path}) is {full_path}"
        )

        return full_path

    def get_directory_from_url(self, url: str, url_parsed: ParseResult) -> str:

        """ This function return directory to write the file. """

        directory = path.normcase(path.dirname(url_parsed.path))

        logging.debug(
            f"Directory for URL: {url} (path: {url_parsed.path}) is {directory}"
        )

        if not directory or directory[0] == "/" or directory[0] == "\\":
            directory = "." + directory
        else:
            directory = path.join(path.normcase("." + self.urlparse.path), directory)

        logging.debug(
            f"Finally directory for URL: {url} (path: {url_parsed.path}) is {directory}"
        )

        return directory


def parse():
    parser = ArgumentParser()
    parser.add_argument(
        "url",
        help="URL to clone.",
    )
    parser.add_argument(
        "--recursive",
        "-r",
        help="Copy data from all urls recursively.",
        action="store_true",
    )
    parser.add_argument(
        "--directory",
        "-d",
        help="Copy data in specific directory.",
        default="CloneWebSite",
    )
    parser.add_argument(
        "--mydomain", "-D", help="Replace domain target by custom domain.", default=None
    )
    parser.add_argument(
        "--myscheme", "-S", help="Replace scheme target by custom scheme.", default=None
    )
    parser.add_argument(
        "--loglevel",
        "-l",
        help="WebSiteCloner logs level.",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "--logfile", "-f", help="WebSiteCloner logs file.", default=None
    )
    return parser.parse_args()


def main():
    args = parse()

    copy: WebSiteCloner = WebSiteCloner(
        args.url,
        recursive=args.recursive,
        directory=args.directory,
        replace_domain=args.mydomain,
        replace_scheme=args.myscheme,
        loglevel=logging.__dict__[args.loglevel],
        logfile=args.logfile,
    )
    copy.launch()


if __name__ == "__main__":
    main()
